/* @(#)fieldef.c	1.38 10/29/02
Copyright 2002 AOI Industries, Inc. All rights reserved.

License is granted to copy, to use, and to make and to use derivative
works for any purpose, provided that the copyright notice and this
license notice is included in all copies and any derivatives works and
in all  related documentation.  Licensor grants no other licenses expressed 
or implied and the licensee acknowleges that Licensor have no liability for 
licensee's use or for any derivative works made by licensee. 

This software is provided AS IS.

Licensor disclaim and licensee agrees that all warranties, express or
implied, including without limitation the implied warranties of
merchantability and fitness for a particular purpose.  Notwithstanding
any other provision contained herein, any liability for damages
resulting from the software or its use is expressly disclaimed,
including consequential or any other indirect damages, whether arising
in contract, tort (including negligence) or strict liability, even if
Licensor is advised of the possibility of such damages.  
 */

/* Parse the field string, build the field record, and write the default
   field, if any.

   The function returns a pointer to the character after the field definition.
   
   Arguments:  fp      - pointer to string containing the field definition,
                         starting at the '%'
	       fldrec  - pointer to the field rec pointer of the caller, which
	                 is filled in after the heap packet is created.
	       row,col - current cursor position.  Note that for list fields,
	                 the actual field start position is col+1
	       scrmode - the character attributes to be set for the field.  For
	                 list fields, the surrounding '[]' are placed in plain
			 mode (which is why the mode is set here, rather than
			 by the caller).
	       ap      - pointer to the argument list of 'putform'.  The
	                 arguments are extracted, and placed in the argument
		         pointer member of the field record, or used as
			 max/min values.
 */
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "form.h"
#include "formdef.h"

#define max(x, y) ((x) > (y) ? (x):(y))


/* Fill the field buffer with blanks */
static void
blankfield(struct FIELDREC *fr)
{
    char *p;
    int   i;

    p = fr->fdata;
    for (i = 0; i < fr->width; i++) *p++ = ' ';

    *p = '\0';
}

char *fieldef(struct FIELD_LIST *fldlist, char *fp, struct FIELDREC **fldrec,
    	      int row, int col, int scrmode, va_list *ap)
{
    char *p, *q, *eofield;
    int noecho, req, def, width, i, precision;
    int minval, maxval;
    struct FIELDREC *fr;
    struct CHOICEREC *cr;
    long            dvl;
    double          dvf;
    enum FSPECS fv;

    p = fp;
    p++;			/* skip past '%' */
    noecho = def = req = FALSE;

    /* Check for no echo, default, read only or required field spec */
    while ((q = strpbrk(p, FIELD_ATTR_SYM)) && p == q) {
	switch (*q) {
	case '?':
	    noecho = TRUE;
	    break;
	case '!':
	    req = TRUE;
	    break;
	case '*':
	    def = TRUE;
	    break;
	}
	p++;
    }

    /* Get field width dimension - please note that width is the width
     * of the field as it is displayed.  Internally the width of the 
     * the allocated list entry must be width+1. For a List field type,
     * automatically calculate width to be accomodate the largest 
     * list entry. 
     */
    q = p;
    if (*q == '>' || *q == '<') {
	/* Allow list field to have 'min' or 'max' range value */
	q++;
	if (*q == '$') q++;
	else while (*q && isdigit(*q)) q++;
    }

    if (*q == '[') {
	width = listlen(q+1);	/* List field - no width allowed */
	precision = -1;
    } else if (*p == '{') {	/* Dir field - no width allowed */
	width = 80;
	precision = -1;
    } else {
	if (sscanf(p, "%d", &width) != 1) width = 0;

	/* check for optional precision */
	while (isdigit(*p)) p++;     /* advance past width */
	if (*p == '.') {
	    if (sscanf(++p, "%d", &precision) != 1) {
		printf("\nfieldef - illegal precision specifier: %*s\n",
		       10 + (p - fp), fp);
		abort();
	    }
          while (isdigit(*p)) p++;     /* advance past precision */
	} else precision = -1;
    }

    if (*p == '(') {	/* Str field - take max of user specified/actual */
	int n = listlen(p+1);
	width = max(n, width);
    }

    if (width < 1) {
	printf("\nfieldef - illegal field width specifier: %*s\n",
	       10 + (p - fp), fp);
	abort();
    }

    /*
     * Create field record, and link it in.  The extra 20 is a fudge factor,
     * and should only need to be 1 (defensive allocation).  The value for
     * field width MUST be computed before this point.
     */
    fr = (struct FIELDREC *)calloc(sizeof(struct FIELDREC) + width + 20, 1);
    if (!fr) {
	printf("\nfieldef: can't allocate memory\n");
	abort();
    }

    *fldrec = fr;
    fr->row = row;
    fr->col = col;
    fr->required = req;
    fr->have_entry = FALSE;
    fr->defval = def;
    fr->echo = !noecho;
    fr->width = width;
    fr->precision = precision;
    fr->smode = scrmode;	/* store attribute mode for field */
    
    /*
     * When the display attribute for the field is unavailable, and no
     * other available attribute is available, the field is written with
     * underscores.  Otherwise, the field is filled with blanks.  The fill
     * character is not stored in the field character buffer - it is only
     * written. 
     */
    if (!scrmode || (*fm_t.us && (fr->smode & UNDERLINE))
	         || (*fm_t.mr && (fr->smode & REVERSE))
	         || (*fm_t.mb && (fr->smode & BLINK))
	         || (*fm_t.md && (fr->smode & BOLD))) {
	fr->fillc = ' ';
    } else {
	fr->fillc = '_';
    }

    /* Append to list of field records */
    fr->flink = NULL;
    if (fldlist->head) {
	fldlist->tail->flink = fr;
	fr->blink = fldlist->tail;
	fldlist->tail = fr;
    } else {
	fldlist->head = fldlist->tail = fr;
	fr->blink = NULL;
    }

    /* find char after field width */
    if (!(frmPeekAhead(&p, RANGE_DATATYPE_SYM, POSITION_SYM))) {
	printf("\nfieldef - illegal field definition: %*s\n", p - fp + 1, fp);
	abort();
    }

    /*
     * Get remainder of field specs - min/max, data type.  If a min/max
     * spec says that the value is in the argument list, it is extracted.
     * When the last spec is found, the pointer to the caller's buffer is
     * placed in the field record. 
     */
    minval = maxval = 0;
    do {
	p = getfspec(fp, p, fr, &minval, &maxval, &fv);
	switch (fv) {
        case INTMIN:
	    fr->dmin = va_arg(*ap, int);
	    break;
        case COORDMIN:
	    fr->fmin = va_arg(*ap, int);
	    break;
        case COORDMAX:
	    fr->fmax = va_arg(*ap, int);
	    break;
	case INTMAX:
	    fr->dmax = va_arg(*ap, int);
	    break;
	case LONGMIN:
	    fr->dmin = va_arg(*ap, long);
	    break;
	case LONGMAX:
	    fr->dmax = va_arg(*ap, long);
	    break;
	case FLOATMIN:
	case DOUBLEMIN:
	    fr->fmin = va_arg(*ap, double);
	    break;
	case FLOATMAX:
	case DOUBLEMAX:
	    fr->fmax = va_arg(*ap, double);
	    break;
	case LASTSPEC:
	    fr->argp = va_arg(*ap, void *);
	    break;
	case NOARGSPEC: break;
	}

    } while (fv != LASTSPEC);

    /* Save return value */
    fp = p;

    if (fr->dtype == List || fr->dtype == Dir || fr->dtype == Str) {
	fr->fillc = ' ';
	fr->required = TRUE;
    }

    /*
     * Now write out field with default field values.  When a default value
     * is supplied that is out of bounds, a field of blanks is displayed. 
     */
    if (fr->defval) {
	switch (fr->dtype) {

    	case Dir:
            /* Trailing blanks are removed, a check is made to remove
               any file extension, the default value is compared against
               allowed choices 
             */
            p = (char *)fr->argp;
            i = strlen(p) - strlen(fr->file_ext);
            if (i > fr->width) i = fr->width;
            q = fr->fdata;
            for (; (i>0) && isprint(*p); i--) *q++ = *p++;
            *q = 0;

            /* scan for match in choice list */
            cr = fr->choice_list;
	    if (cr) {
		do {
		    if (!strcmp(cr->choice, fr->fdata)) {
			fr->curr_choice = cr;
			cr = 0;
			break;
		    }
		    cr = cr->next;
		} while (cr != fr->choice_list);
	    }

            if (cr) { 
		/* just pick up first entry */
		strcpy(fr->fdata, cr->choice);
		for (i=strlen(fr->fdata); i<fr->width; i++)
		    fr->fdata[i] = fr->fillc;

		fr->fdata[fr->width] = 0;
    	    	fr->curr_choice = cr;
	    }
	    fr->choice_displayed = TRUE;
            break;

	case Char:
	    /*
	     * Copy out default string data, left-justified and blank
	     * padded. 
	     */
	    p = (char *)fr->argp;
	    while (*p && isspace(*p)) p++;

	    /*
	     * Any non-blank character in field qualifies as an entry. 
	     */
	    if (*p)
		fr->have_entry = TRUE;
	    else
		fr->have_entry = FALSE;

            q = fr->fdata;
            eofield = &q[fr->width];
            while (q < eofield && *p && isprint(*p)) *q++ = *p++;
            
            /* blank pad */
            while (q < eofield) *q++ = ' ';
            *q = 0;
	    break;

	case Int:
    	    if (precision > width)
		fr->width = precision;

    	    if (!minval) fr->dmin = field_minint(fr->width);
    	    if (!maxval) fr->dmax = field_maxint(fr->width);

	    dvl = *(int *) fr->argp;
	    if ((dvl < fr->dmin) || (dvl > fr->dmax)) {
		blankfield(fr);
		fr->have_entry = FALSE;
	    } else {
		if (fr->precision >= 0)
		   sprintf(fr->fdata, "%-*.*d", fr->width,
                           fr->precision, (int) dvl);
		else
		   sprintf(fr->fdata, "%-*d", fr->width, (int) dvl);

		fr->have_entry = TRUE;
	    }
	    break;

	case Long:
    	    if (precision > width)
		fr->width = precision;

    	    if (!minval) fr->dmin = field_minint(fr->width);
    	    if (!maxval) fr->dmax = field_maxint(fr->width);

	    dvl = *(long *) fr->argp;
	    if ((dvl < fr->dmin) || (dvl > fr->dmax)) {
		blankfield(fr);
		fr->have_entry = FALSE;
	    } else {
		if (fr->precision >= 0)
		   sprintf(fr->fdata, "%-*.*ld", fr->width, fr->precision,dvl);
                else
                   sprintf(fr->fdata, "%-*ld", fr->width, dvl);

		fr->have_entry = TRUE;
	    }
	    break;

	case Hex:
    	    if (precision > width)
		fr->width = precision;

    	    if (!minval) fr->dmin = minval;
    	    if (!maxval) fr->dmax = ULONG_MAX;

	    dvl = *(unsigned int *) fr->argp;
	    if (((unsigned int) dvl < (unsigned int) fr->dmin)
		|| ((unsigned int) dvl > (unsigned int) fr->dmax)) {

		blankfield(fr);
		fr->have_entry = FALSE;
	    } else {
                if (fr->precision >= 0)
                   sprintf(fr->fdata, "%*.*X", fr->width,
                          fr->precision, (unsigned int) dvl);
                else
                   sprintf(fr->fdata, "%*X", fr->width, (unsigned int) dvl);
		fr->have_entry = TRUE;
	    }
	    break;

	case Float:
	case Double:
    	    if (precision + 3 > width)
		fr->width = precision + 3;

    	    if (!minval) fr->fmin = field_minflt(fr->width, fr->precision);
    	    if (!maxval) fr->fmax = field_maxflt(fr->width, fr->precision);

    	    if (fr->precision < 0) {
	        printf("\nfieldef - float/double field requires\
 precision: %*s\n", p - fp + 1, fp);
	    	abort();
	    }

	    if (fr->dtype == Float) dvf = *(float *) fr->argp;
	    else                    dvf = *(double *) fr->argp;

	    if ((dvf < fr->fmin) || (dvf > fr->fmax)) {
		blankfield(fr);
		fr->have_entry = FALSE;
	    } else {
		sprintf(fr->fdata, "%-*.*f",fr->width,fr->precision, dvf);
		fr->have_entry = TRUE;
	    }
	    break;

    	case List:
	    dvl = *(unsigned int *) fr->argp;
	    if (maxval) {
		if ((1 << dvl) > fr->dmax) dvl = 0;
	    } else if ((unsigned int)dvl > (fr->num_choices-1)) dvl = 0;
	
	    for (cr=fr->choice_list;
		 cr->ord != (unsigned int)dvl
		     && cr->next != fr->choice_list; cr = cr->next) ;

	    strcpy(fr->fdata, cr->choice);
	    for (i=strlen(fr->fdata); i<fr->width; i++) fr->fdata[i] = ' ';
	    
	    fr->fdata[fr->width] = 0;
	    fr->curr_choice = cr;
	    fr->choice_displayed = TRUE;
	    break;

	case Str:
	    strcpy(fr->fdata, (char *)fr->argp);
	    to_caps(fr->fdata);
	    for (i=strlen(fr->fdata); i<fr->width; i++) fr->fdata[i] = ' ';
	    fr->fdata[fr->width] = 0;
	    fr->choice_displayed = FALSE;
	    break;
	case None:
	}
    } else {
	/* No default value supplied */
	blankfield(fr);
    	switch (fr->dtype) {
	  case  Long:
	  case  Int:
	    if (!minval) fr->dmin = field_minint(fr->width);
	    if (!maxval) fr->dmax = field_maxint(fr->width);
	    break;
	  case  Hex:
	    if (!minval) fr->dmin = 0;
	    if (!maxval) fr->dmax = ULONG_MAX;
	    break;
	  case Float:
	  case Double:
	    if (!minval) fr->fmin = field_minflt(fr->width, fr->precision);
	    if (!maxval) fr->fmax = field_maxflt(fr->width, fr->precision);
	    break;
	    
	  case Str:
	  case List:
	  case Dir:
	    /* just pick up the first element on the list */
	    fr->curr_choice = fr->choice_list;
	    strcpy(fr->fdata, fr->choice_list->choice);
	    for (i=strlen(fr->fdata); i<fr->width; i++)
		fr->fdata[i] = fr->fillc;
	    fr->fdata[fr->width] = 0;
	    fr->choice_displayed = TRUE;
	    break;
	    
	  case None:
	  case Char:
	}
    }

    /* Write out the field data, using the fill character if necessary. */
    if (fr->dtype == List || fr->dtype == Dir || fr->dtype == Str) {
        fr->col++;       /* make room for initial '[' */ 
        if (scrmode) reset_attr(0);
	putchar('[');
    }

    reset_attr(scrmode);
    width = strlen(fr->fdata);
    for (i=0; i<width; i++)
	if (isspace(fr->fdata[i]))
	    putchar(fr->fillc);
	else
	    putchar(fr->fdata[i]);
	    
    for (i=width; i<fr->width; i++) putchar(fr->fillc);

    if (fr->dtype == List || fr->dtype == Dir || fr->dtype == Str) {
	/* Set screen attribute back to normal */
    	reset_attr(0);
	putchar(']');
    	reset_attr(scrmode);
    }

    return fp;
}
