/* @(#)getfspec.c	1.24 10/29/02
Copyright 2002 AOI Industries, Inc. All rights reserved.

License is granted to copy, to use, and to make and to use derivative
works for any purpose, provided that the copyright notice and this
license notice is included in all copies and any derivatives works and
in all  related documentation.  Licensor grants no other licenses expressed 
or implied and the licensee acknowleges that Licensor have no liability for 
licensee's use or for any derivative works made by licensee. 

This software is provided AS IS.

Licensor disclaim and licensee agrees that all warranties, express or
implied, including without limitation the implied warranties of
merchantability and fitness for a particular purpose.  Notwithstanding
any other provision contained herein, any liability for damages
resulting from the software or its use is expressly disclaimed,
including consequential or any other indirect damages, whether arising
in contract, tort (including negligence) or strict liability, even if
Licensor is advised of the possibility of such damages.  
 */
/* Interpret the field spec pointed to by 'sp', assumed to be one of
   ".<>sdlxfFXF", setting the appropriate value in the field rec pointed 
   to by 'fr'.  

   The function returns 
   	LASTSPEC when a field data type spec is read (last spec!)
	NOARGSPEC field with no max/min argument
	INTMAX int max field with argument value
	INTMIN int min field with argument value
        LONGMAX long max field with argument value
        LONGMIN long min field with argument value
        FLOATMAX float max field with argument value
        FLOATMIN float min field with argument value
        DOUBLEMAX double max field with argument value
        DOUBLEMIN double min field with argument value

   The pointer 'sp' is advanced to the character after the spec.

   The pointer 'fp' points to the start of the field definition, and is used
   for error messages.

   Errata:  The parsing of the directory and list fields really doesn't
            belong in this routine....
 */
#define _BSD_SOURCE
#define __EXTENSIONS__
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <dirent.h>
#include "form.h"
#include "formdef.h"

/* Check to see if the min/max value is given by an argument to 'putform'
 */

static int argvalue(c)
   char **c;
{
    if (**c == '$') {
	(*c)++;
	return TRUE;
    } else
	return FALSE;
}

char *
getfspec(char *fp, char *sp, struct FIELDREC *fr, int *minval, int *maxval,
	 enum FSPECS *fval)
{
    int   s, ord, extlen, micron, item_cnt, selectMask;
    char *p, *env, dirname[128], envname[128];
    DIR  *dirp;
    struct dirent    *dp;
    struct CHOICEREC *cur, *last;
    struct CHOICEREC *cr, *ncr;

    switch (*sp) {

      case  '{':

    	if (*sp++ == '}') {	    	/* empty choice list */
    	    printf("\ngetfspec - Empty directory list not allowed\n");
	    exit(-1);
	}

        /* Get the directory name.  Inclusion of a single environment variable
	   at the start of the name and a single wildcard specification to
	   to specify a required file extension is allowed.
	 */
    	p = dirname;
	*p = 0;
    	if (*sp == '$') { /* Start of an environment variable */
    	    sp++; /* Skip past the '$' */

    	    /* Accumulate the environment variable */
	    env = envname;
    	    while (*sp && (*sp != '/') && (*sp != '}')) *env++ = *sp++;
	    *env = 0;

            /* Get the environment variable */
    	    if (!(p = getenv(envname))) {
		printf("\ngetfspec:No environment variable set for %s\n",
		       envname);
		exit(-1);
	    }
    	    strcpy(dirname, p);
	    p = &dirname[strlen(dirname)];
    	}

	/* copy over remainder of pathname, checking for a file extension. */
	fr->file_ext[0] = 0;
    	while (*sp && (*sp != '}')) {
	    if (*sp == '*') {	/* check for file extension */
		*(p-1) = 0;	/* terminate dirname, eliminating '/' */
		sp++;		/* skip '*', leaving sp point to initial '.' */
		/* change pointer to copy remainder of string to file_ext */
		p = fr->file_ext;
	    } else 
		*p++ = *sp++;
	}
	*p = 0;

        /*
         * now we have the directory name, lets use the
         * directory commands to get each entry one at a
         * time and then stuff the name into a choice record.
         */
    	if (!(dirp = opendir(dirname)))	{	    
	    logError(errno, "Can't open %s", dirname);
	}

        cur = NULL;  last = NULL;  fr->choice_list = NULL;
	ord = 0;
	extlen = strlen(fr->file_ext);
	fr->width = 0;

        /*
         * now we have the directory name, lets use the
         * directory commands to get each entry one at a
         * time and then stuff the name into a choice record.
         */
    	if (!(dirp = opendir(dirname)))	{
	    logError(errno, "Can't open %s", dirname);
	    break;
	}

	while ((dp = readdir(dirp))) {
	    if (dp->d_name[0] == '.') continue;

	    /* Check for match against file extension */
	    s = strlen(dp->d_name) - extlen;
	    if ((extlen && (s <= 0))
		|| strcmp(&dp->d_name[s], fr->file_ext)) continue;

            dp->d_name[s] = 0;  /* terminate after extension */
	    if (s > fr->width) fr->width = s;

            if (!(cur = malloc(sizeof(struct CHOICEREC)))) {
                printf("\ngetfspec - Cannot allocate choice record\n");
                exit(-1);
            }

            cur->next = NULL;
            cur->prev = last;

            cur->ord  = ord++;
            cur->cand = FALSE;

	    if ((cur->choice = strdup(dp->d_name)) == NULL) {
	    	printf("\ngetfspec - Cannot allocate choice\n");
	    	exit(-1);
	    }  /* if malloc failed */

            /* Set the beginning of the list up. */
            if (fr->choice_list == NULL) fr->choice_list = cur;

            /* make sure to add connection from last to cur before moving on */
            if (last) last->next = cur;

            last = cur;
	}

    	closedir(dirp);

    	if (!(fr->curr_choice = fr->choice_list)) {
	    logError(0, "Empty directory list: %s", dirname);
	}

    	fr->num_choices = ord;

    	if (fr->choice_list) {
	    last->next = fr->choice_list;
	    fr->choice_list->prev = last;
	}
	*fval = NOARGSPEC;
        break;  /* directory fields */

      case  '(':
      case  '[':	    	    	/* list begin */
        /* go past '[' */
        sp++;

        /* set things up so pointers can be set properly */
        cr = fr->choice_list = NULL;
	ord = 0;
	item_cnt = 0;
	selectMask = 0;
	if (fr->dmax) selectMask = fr->dmax;
	if (fr->dmin) selectMask = fr->dmin;

        for (;*sp != ']' && *sp != ')';) {
            /* Get the next choice in the string.  When a "minimum" or
	       "maxium" value was defined, use it as a bit field to determine
	       which list elements to keep.  e.g., for a value of 5,
	       retain the first and third elements, and drop the second.

	       For a 'min' value, the ordinal of the field corresponds to
	       the number of fields selected.

	       For a 'max' value, the ordinal of the field corresponds to
	       the position in the original 'raw' list of fields.
	    */
	    for (p=0; *sp != ']';) {

		p = next_choice(&sp, &s);
		if (selectMask && !(selectMask & (1 << item_cnt))) {
		    /* skip this entry */
		    item_cnt++;
		    p = 0;
		    if (*sp == ',') sp++;
		} else break;	/* have selection */
	    }

	    if (p) {

		if (!(ncr = calloc(sizeof(struct CHOICEREC), 1))) {
		    printf("\ngetfspec - Cannot allocate choice record\n");
		    exit(-1);
		}

		if (fr->dmin)
		    ncr->ord = ord;
		else
		    ncr->ord = item_cnt;

		ord++;
		item_cnt++;

		/* If this is first, set fr->choice_list,
		   else tack on to list.
		*/
		if (fr->choice_list == NULL) {
		    fr->choice_list = ncr;
		    ncr->prev = NULL;
		} else {
		    cr->next = ncr;
		    ncr->prev = cr;
		}
		
		/* malloc space to hold choice string. */
		if ((ncr->choice = strdup(p)) == NULL) {
		    printf("\n\ngetfspec: can not allocate memory for \
choice string: %s\n", p);
		    exit(-1);
		}
		
		to_caps(ncr->choice);
		
		/* remember next_choice leaves sp @ ',' or  `]' */
		if (*sp == ',') sp++;

		cr = ncr;
	    }

        }  /* while choices left in string */

        /* make the list circular */
        cr->next = fr->choice_list;
        fr->choice_list->prev = cr;

        fr->num_choices = ord;
        fr->curr_choice = cr;
	*fval = NOARGSPEC;
	break;			/* list fields */

    case '<':			/* Maximum */
	*maxval = TRUE;
	sp++;
    	    /* peek ahead for data type */
    	p = sp;
	if (!(frmPeekAhead(&p, DATATYPE_SYM, POSITION_SYM))) {
	    printf("\ngetfspec - field data type not found: %*s\n",
		   10 + (sp - fp), fp);
	    exit(-1);
	}

	switch (*p) {

	case 'b':
	case 's':
	  printf("\ngetfspec: string/barcode spec can't have max value: %*s\n",
		 10 + (sp - fp), fp);
	  exit(-1);

	case 'x':
	case 'd':
	case '[':
	    if (argvalue(&sp)) {
		*fval = INTMAX;
		return sp;
	    }
	    /* no break intended */

	case 'l':
	    if (argvalue(&sp)) {
		*fval = LONGMAX;
		return sp;
	    }

	    s = sscanf(sp, "%ld", &(fr->dmax));
	    if (s != 1) {
		printf("\ngetfspec - illegal max specifier: %*s\n",
		       10 + (sp - fp), fp);
		exit(-1);
	    }
	    break;

	case 'X':
	case 'Y':
	    if (argvalue(&sp)) {
		*fval = COORDMAX;
		return sp;
	    }

	    s = sscanf(sp, "%d", &micron);
	    if (s != 1) {
		printf("\ngetfspec - illegal max specifier: %*s\n",
		       10 + (sp - fp), fp);
		exit(-1);
	    }
            fr->fmax = micron;
	    break;


	case 'f':
	    if (argvalue(&sp)) {
		*fval = FLOATMAX;
		return sp;
	    }
	    /* no break intended */

	case 'F':
	    if (argvalue(&sp)) {
		*fval = DOUBLEMAX;
		return sp;
	    }

	    s = sscanf(sp, "%lf", &(fr->fmax));
	    if (s != 1) {
		printf("\ngetfspec - illegal max specifier: %*s\n",
		       10 + (sp - fp), fp);
		exit(-1);
	    }
	    break;
	}
	sp = strpbrk(sp, RANGE_DATATYPE_SYM);
	*fval = NOARGSPEC;
	break;			/* Maximum */

    case '>':			/* Minimum */
	*minval = TRUE;
	sp++;
    	    /* peek ahead for data type */
    	p = sp;
	if (!(frmPeekAhead(&p, DATATYPE_SYM, POSITION_SYM))) {
	    printf("\ngetfspec - field data type not found: %*s\n",
		   10 + (sp - fp), fp);
	    exit(-1);
	}

	switch (*p) {
	case 'b':
	case 's':
	 printf("\ngetfspec - string/barcode spec can't have min value: %*s\n",
		10 + (sp - fp), fp);
	 exit(-1);

	case 'x':
	case 'd':
	case '[':
	    if (argvalue(&sp)) {
		*fval = INTMIN;
		return sp;
	    }
	    /* no break intended */

	case 'l':
	    if (argvalue(&sp)) {
		*fval = LONGMIN;
		return sp;
	    }

	    s = sscanf(sp, "%ld", &(fr->dmin));
	    if (s != 1) {
		printf("\ngetfspec - illegal min specifier: %*s\n",
		       10 + (sp - fp), fp);
		exit(-1);
	    }
	    break;

	case 'f':
	    if (argvalue(&sp)) {
		*fval = FLOATMIN;
		return sp;
	    }
	    /* no break intended */

	case 'F':
	    if (argvalue(&sp)) {
		*fval = DOUBLEMIN;
		return sp;
	    }

	    s = sscanf(sp, "%lf", &(fr->fmin));
	    if (s != 1) {
		printf("\ngetfspec - illegal min specifier: %*s\n",
		       10 + (sp - fp), fp);
		exit(-1);
	    }
	    break;
	}
	sp = strpbrk(sp, RANGE_DATATYPE_SYM);
	*fval = NOARGSPEC;
	break;			/* Minimum */

    case 's':			/* String Data Type */
	fr->dtype = Char;
	sp++;
	*fval = LASTSPEC;
	return sp;

    case 'd':			/* Integer */
	fr->dtype = Int;
	sp++;
	*fval = LASTSPEC;
	break;

    case 'l':			/* Long */
	fr->dtype = Long;
	sp++;
	*fval = LASTSPEC;
	break;

    case 'x':			/* Hex Int */
	fr->dtype = Hex;
	sp++;
	*fval = LASTSPEC;
	break;

    case 'f':			/* Float */
	fr->dtype = Float;
	sp++;
	*fval = LASTSPEC;
	break;

    case 'F':			/* Double */
	fr->dtype = Double;
	sp++;
	*fval = LASTSPEC;
	break;

    case  '}':	    	    	/* Directory list */
    	fr->dtype = Dir;
	sp++;
	*fval = LASTSPEC;
	break;

    case  ']':	    	    	/* List */
    	fr->dtype = List;
	sp++;
	*fval = LASTSPEC;
	break;

    case  ')':	    	    	/* String */
    	fr->dtype = Str;
	sp++;
	*fval = LASTSPEC;
	break;

    default:
	printf("\ngetfspec - illegal specifier in field: %.*s\n",
	       1 + (sp - fp), fp);
	exit(-1);
    }

    return sp;
}
