 /* @(#)initform.c	1.29 6/13/00
Copyright 2002 AOI Industries, Inc. All rights reserved.

License is granted to copy, to use, and to make and to use derivative
works for any purpose, provided that the copyright notice and this
license notice is included in all copies and any derivatives works and
in all  related documentation.  Licensor grants no other licenses expressed 
or implied and the licensee acknowleges that Licensor have no liability for 
licensee's use or for any derivative works made by licensee. 

This software is provided AS IS.

Licensor disclaim and licensee agrees that all warranties, express or
implied, including without limitation the implied warranties of
merchantability and fitness for a particular purpose.  Notwithstanding
any other provision contained herein, any liability for damages
resulting from the software or its use is expressly disclaimed,
including consequential or any other indirect damages, whether arising
in contract, tort (including negligence) or strict liability, even if
Licensor is advised of the possibility of such damages.  
 */

/* Perform initial setup of the 'forms' package.  The standard
   input file (assumed to be a tty device) with echo off, and put in CBREAK or
   RAW mode, with buffered IO.

    	initform(mode, pc_input, locale, procID)

    	    int mode     - 0 : CBREAK
    	    	    	   1 : RAW
    	    	    	   2 : CBREAK with block buffered mode
			   3 : RAW with block buffered mode

    	    int pc_input - query the ER station for coordinate fields.  Assumes
    	    	           that a call has been made to 'vrio' to initialize
    	    	           communications.
    	    char *locale - current language context
    	    int procID   - If non-zero, communications will be established with
    	    	           the logserver with a logUIInit(LOCAL_HOST, procID).

   Exit and SIGINT/SIGTERM handlers are set up to put the terminal back the
   way it was on entry.
   
   The termcap file for the current environment is processed.

   NOTE:  Buffered IO mode is turned on, so output is not written until an 
          fflush(stdout) is called.  Normally, an operator interface routine
          will be waiting in 'getKchar' for a character, so a call is made
          there prior to suspending on a character.

          A call to fflush(stdout) will need to be made at any other point
          that the code suspends, if output should be displayed.

    	  This minimizes ethernet traffic if the display is being used over
    	  a network connection.
 */
#ifdef __linux__
#define _POSIX_SOURCE
#define _BSD_SOURCE
#endif
#define __EXTENSIONS__
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <termio.h>
#if defined(__sun__) && defined(__svr4__)
#include <curses.h>
#include <term.h>
#endif

#include <form.h>
#include <formdef.h>

struct TERMTYPE fm_t;

int default_language;	/* TRUE if form/menu starts with ` */
static struct termios std_setup;

/* For some reason, the kP and kN definitions get filtered out on SunOS;
   put them in here for our common terminals.
 */
enum TERMNAME { xterm, vt200, MAXTERMS };
static struct {
    char *kP;
    char *kN;
} termcap[] = { { "\033[5z", "\033[6z" },
	      { "\033[5~", "\033[6~" },
};

/* Pick up the termcap info and store it */
#define get_termcap(str, dest) if (!(dest = tgetstr(str, &p))) dest = ""

static void form_sig_handler() { exit(0); }

static int fNorm = 1;

void uiSetNorm(void)
{
    tcsetattr(fileno(stdin), TCSANOW, &std_setup);
    normal_video();		/* special modes off */
    writeterm(fm_t.se);	/* standout mode off */
    cursor_on();		/* cursor on */
}

/******************************************************************************
* FUNCTION NAME: uiSetForm
*
* DESCRIPTION: Sets stdin and stdout to form mode (i.e. no ECHO, CBREAK, and
*   	       optional RAW).
*	       
* RETURNS: 0 on success, -1 on failure with errno set by failed system call.
*****************************************************************************/
int uiSetForm(int mode)
{
    int in;
    struct termios fsetup;
    FILE *new;
    
    /* See if we're already in form mode */
    if (!fNorm) return (0);

    fNorm = 0;

    in = fileno(stdin);
    tcgetattr(in, &std_setup);

    fsetup = std_setup;

    fsetup.c_iflag = IGNBRK|ISTRIP|((IXON|IXANY|IXOFF)&std_setup.c_iflag);
    fsetup.c_oflag = 0;
    fsetup.c_cflag = CS8|CREAD|(std_setup.c_cflag & CBAUD)
			|((CRTSCTS|CLOCAL|HUPCL)&std_setup.c_cflag);

    if (mode & TTY_RAW_MODE)
	fsetup.c_lflag = 0;
    else
	fsetup.c_lflag = ISIG;
    
    fsetup.c_cc[VMIN] = 1;
    fsetup.c_cc[VTIME] = 0;

    tcsetattr(in, TCSANOW, &fsetup);

    /* Set block buffered mode */
    if (mode & TTY_BLOCK_MODE) {
    	if (!(new = freopen(ttyname(fileno(stdout)), "w", stdout))) {
	    perror("uiSetForm - Cannot reopen stdout");
	    return (-1);
    	}

    	if (setvbuf(stdout, NULL, _IOFBF, BUFSIZ) < 0) {
	    perror("Set buffered IO mode error");
	    return (-1);
    	}
    }
    
    return (0);
}

void initform(int mode)
{
    int s;
    char bp[1024], *termname, *p, *textdomain();
    enum TERMNAME term;

    if (uiSetForm(mode) < 0) exit(-1);

    /* Read interesting terminal characteristics */
    termname = getenv("TERM");
    if (!termname) {
        printf("environment variable TERM undefined\n");
        exit(-1);
    }
	
    s = tgetent(bp, termname);	/* read termcap file */
	
    if (s == -1) {
        perror("Cannot open TERMCAP file");
        exit(-1);
    }
    if (s == 0) {
        printf("No termcap entry for %s\n", termname);
        exit(-1);
    }
	
    /* Retrieve terminal command sequences.  There appears to be a
       mistake in the SUN termcap entry for the vt100.  The sequence
       described for the arrow keys is not what is actually sent, but
       'vi' requires those definitions.  They are converted here to the
       actual sequence sent by a VT100.
     */
    p = fm_t.tbuf;

    get_termcap("cl", fm_t.cl);
    if (!*fm_t.cl) { puts("tty doesn't have cl function"); exit(-1); }
    get_termcap("ce", fm_t.ce);
    if (!*fm_t.ce) { puts("tty doesn't have ce function"); exit(-1); }
    get_termcap("cd", fm_t.cd);
    if (!*fm_t.cd) { puts("tty doesn't have cd function"); exit(-1); }
    get_termcap("cm", fm_t.cm);
    if (!*fm_t.cm) { puts("tty doesn't have cm function"); exit(-1); }
    get_termcap("kd", fm_t.kd);
    if (!*fm_t.kd) { puts("tty doesn't have kd function"); exit(-1); }
    if (!strncmp(fm_t.kd, "\033O", 2)) fm_t.kd[1] = '[';
    get_termcap("kl", fm_t.kl);
    if (!*fm_t.kl) { puts("tty doesn't have kl function"); exit(-1); }
    if (!strncmp(fm_t.kl, "\033O", 2)) fm_t.kl[1] = '[';
    get_termcap("kr", fm_t.kr);
    if (!*fm_t.kr) { puts("tty doesn't have kr function"); exit(-1); }
    if (!strncmp(fm_t.kr, "\033O", 2)) fm_t.kr[1] = '[';
    get_termcap("ku", fm_t.ku);
    if (!*fm_t.ku) { puts("tty doesn't have ku function"); exit(-1); }
    if (!strncmp(fm_t.ku, "\033O", 2)) fm_t.ku[1] = '[';
    get_termcap("k1", fm_t.k1);
    if (!*fm_t.k1) { puts("tty doesn't have k1 function"); exit(-1); }
    get_termcap("k2", fm_t.k2);
    if (!*fm_t.k2) { puts("tty doesn't have k2 function"); exit(-1); }
    get_termcap("k3", fm_t.k3);
    if (!*fm_t.k3) { puts("tty doesn't have k3 function"); exit(-1); }
    get_termcap("k4", fm_t.k4);
    if (!*fm_t.k4) { puts("tty doesn't have k4 function"); exit(-1); }
    get_termcap("kh", fm_t.kh);
    if (!*fm_t.kh) { puts("tty doesn't have home function"); exit(-1); }
    get_termcap("kC", fm_t.kc);
    get_termcap("mb", fm_t.mb);
    get_termcap("md", fm_t.md);
    get_termcap("mr", fm_t.mr);
    get_termcap("me", fm_t.me);
    get_termcap("nd", fm_t.nd);
    if (!*fm_t.nd) { puts("tty doesn't have nd function"); exit(-1); }
    get_termcap("WE", fm_t.WE);
    get_termcap("WS", fm_t.WS);
    get_termcap("ue", fm_t.ue);
    get_termcap("us", fm_t.us);
    get_termcap("so", fm_t.so);
    get_termcap("se", fm_t.se);
    get_termcap("dl", fm_t.dl);
    get_termcap("CE", fm_t.CE);
    get_termcap("CS", fm_t.CS);
    get_termcap("kN", fm_t.kN);
    get_termcap("kP", fm_t.kP);
    get_termcap("%1", fm_t.help);

    /* substitute for handicapped 'termcap's */
    if (!strcmp(termname, "xterm"))
	term = xterm;
    else if (!strcmp(termname, "vt200"))
	term = vt200;
    else 
	term = MAXTERMS;

    if (term < MAXTERMS) {
	if (! *fm_t.kP) fm_t.kP = termcap[term].kP;
	if (! *fm_t.kN) fm_t.kN = termcap[term].kN;
    }

    atexit(uiSetNorm);

    signal(SIGINT, form_sig_handler);
    signal(SIGTERM, form_sig_handler);

    clr_screen();
}



