/* @(#)readonly.c	1.16 10/24/00
Copyright 2002 AOI Industries, Inc. All rights reserved.

License is granted to copy, to use, and to make and to use derivative
works for any purpose, provided that the copyright notice and this
license notice is included in all copies and any derivatives works and
in all  related documentation.  Licensor grants no other licenses expressed 
or implied and the licensee acknowleges that Licensor have no liability for 
licensee's use or for any derivative works made by licensee. 

This software is provided AS IS.

Licensor disclaim and licensee agrees that all warranties, express or
implied, including without limitation the implied warranties of
merchantability and fitness for a particular purpose.  Notwithstanding
any other provision contained herein, any liability for damages
resulting from the software or its use is expressly disclaimed,
including consequential or any other indirect damages, whether arising
in contract, tort (including negligence) or strict liability, even if
Licensor is advised of the possibility of such damages.  
 */
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "formdef.h"
#include "form.h"

#define BUF_SIZE 80

/*
 * Function: readonly(fp, scrmode, col, ap)
 *
 * Description:
 *      Handle readonly fields w/o treating it as a real field. IE - don't
 *      malloc space for it, and don't add it to the list of fields.  One
 *      drawback to this is we interperet read only fields differently from
 *      all other fields.  This leads to the non-ability to specify minimum
 *      and maximum values for them (one of the most significant differences
 *      between these and normal fields).  One benefit is that any sprintf
 *      format string is acceptable (IE you can have variable length read only
 *      fields, one of the other more significant differences between the two
 *      field types).
 *
 * Parameters:
 *      fp      - ptr to string containing form specification (starting at "%")
 *      scrmode - character attributes to be set for field.
 *      col     - ptr to current column.
 *      ap      - ptr to argument list of 'putform' call.
 *
 * Returns:
 *      pointer to form specification after readonly field
 *
 * Note:
 *      Read ony directory fields are NOT supported.
 *
 */

char *readonly(char *fp, int scrmode, int *col, va_list *ap)
{
    char *p, *q, *z;
    char tmp[200];
    char *ret;
    int i, len, item_num, tmplen;
    char buf[BUF_SIZE];

    p = fp+2; /* skip past '%' and '#' */
    memset(buf, '\0', BUF_SIZE);
    ret = NULL;

    if ((q = strpbrk(p, DATATYPE_SYM)) == NULL) {
        printf ("\nIllegal read only field specification, %s\n", fp);
        exit (-1);
    }

    /*
     * Once we know what the data type for the field is, we can get things
     * set up to convert data to a displayable string.
     */
    switch(*q) {
        case  '{':
            printf("\nDirectory fields can not be read only!: %s\n", fp);
            exit (-1);

        case  '[':
            z = q + 1;
            i = 0;

            /* figure out which item number is the default */
            item_num = *(va_arg(*ap, int *));

            /* Figure out the maximum length of the list field entries */
            len = listlen(z);

            /* Skip down the list of items, until we've gotten 'item_num' */
            if (item_num >= 0) {
                for (; i <= item_num && *z != ']'; i++) {
                    p = next_choice(&z, &tmplen);

                    /* skip over the ',' */
                    if (*z == ',') z++;
                }
            }

            /* If item number is out of range just pad w/ blanks */
            if (item_num < 0 || i -1 != item_num)
                buf[0] = '\0';

            else strcpy(buf, p);

            /* blank pad takes into account '\0' */
            blankpad(buf, len + 1);

            /*
             * For all other fields this gets set below.  Return value
             * is end of field + 1
             */
            ret = strpbrk(z, "]") + 1;
            break;

        default:
            /*
             * For all fields that aren't list fields, we just copy the format
             * specifier, then use that for a sprintf call later on.
             */

            /*
             * since q is at last character of field definition, we should take
             * advantage of this and set the return value
             */
            ret = q + 1;
            len = q - p + 1;

            memset(tmp, '\0', len+4);
    
            /* set up the string to be sent to printf */
            *tmp = '%';

	    strncpy(tmp+1, p, len);

            /* check and make sure no other attributes specified */
            if ((z = strpbrk(tmp, "?!*")) != NULL) {
                printf("\nAttribute \"%c\" not allowed for read only fields: %s\n",
                           *z, fp);
                exit (-1);
            }

	    q = tmp + len;

            /* Depending on data type, convert the field. */
            switch (*q) {
                case 's':
            	    *(tmp+1) = '-'; 	    	/* left justify */
            	    strncpy(tmp+2, p, len);	/* +2 so we skip % and - */
                    sprintf(buf, tmp, va_arg(*ap, char *));
                    break;

                case 'd':
                    sprintf(buf, tmp, *(va_arg(*ap, int *)));
                    break;
 
                case 'l':
                    *(tmp+len+1) = 'd';
                    sprintf(buf, tmp, *(va_arg(*ap, long *)));
                    break;
                
                case 'x':
            	    *(tmp+1) = '0'; 	    	/* zero fill */
            	    strncpy(tmp+2, p, len);	/* +2 so we skip % and - */
                    *(tmp+len+1) = 'X';
                    sprintf(buf, tmp, *(va_arg(*ap, unsigned int *)));
                    break;

                case 'F':
    	    	    *(tmp+len) = 'f';
                    sprintf(buf, tmp, *(va_arg(*ap, double *)));
                    break;

                case 'f':
                    sprintf(buf, tmp, *(va_arg(*ap, float *)));
                    break;

                default:
                    printf("\nField Type \"%c\" invalid for read only fields: %s\n",
                           *q, fp);
                    exit (-1);
                    break;
            }  /* switch all field types but list */
    }  /* switch field type */

    if (*q == '[') {
        if (scrmode) reset_attr(0);
        putchar('[');
    }

    reset_attr(scrmode);
    fputs(buf, stdout);

    if (*q == '[') {
        if (scrmode) reset_attr(0);
        putchar(']');
        reset_attr(scrmode);
    }

    *col += strlen(buf) + (*q == '[' ? 2 : 0);

    return (ret);
}

