/* @(#)valid_field.c	1.24 10/29/02
Copyright 2002 AOI Industries, Inc. All rights reserved.

License is granted to copy, to use, and to make and to use derivative
works for any purpose, provided that the copyright notice and this
license notice is included in all copies and any derivatives works and
in all  related documentation.  Licensor grants no other licenses expressed 
or implied and the licensee acknowleges that Licensor have no liability for 
licensee's use or for any derivative works made by licensee. 

This software is provided AS IS.

Licensor disclaim and licensee agrees that all warranties, express or
implied, including without limitation the implied warranties of
merchantability and fitness for a particular purpose.  Notwithstanding
any other provision contained herein, any liability for damages
resulting from the software or its use is expressly disclaimed,
including consequential or any other indirect damages, whether arising
in contract, tort (including negligence) or strict liability, even if
Licensor is advised of the possibility of such damages.  
*/

/*
 * Validate the contents of the field, and copy the value back to the
 * place specified by the caller of 'putform', if the 'ret' flag is true.
 *
 * Character data must only be printable, and that's been screened for earlier.
 * Integer/long data must not have a period, only one leading plus or minus
 * sign, and all digits.  Float/double are like integers, but can have a '.'.
 * Arguments for non-character fields are filled in with a special marker
 * when the fields were left blank.
 *
 * Returns:     0 - field not valid
 * 		1 - field valid
 */
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "form.h"
#include "formdef.h"

#define irint(x)   ((int)((x) > 0 ? ((x) + .5):((x) - .5)))
#define max(x, y) ((x) > (y) ? (x):(y))

static
void badmin_int(struct FIELDREC *cf)
{
    logError(0, "Value must be at least %ld\7", cf->dmin);
}

static
void badmax_int(struct FIELDREC *cf)
{
    logError(0, "Value must be at most %ld\7", cf->dmax);
}
static
void badmin_float(struct FIELDREC *cf)
{
    double v;

    /* for coord fields, display units are mm, but internally stored
       as microns.
     */
    v = cf->fmin;

    logError(0, "Value must be at least %*.*f\7", cf->width, cf->precision, v);
}

static
void badmax_float(struct FIELDREC *cf)
{
    double v;

    /* for coord fields, display units are mm, but internally stored
       as microns.
     */
    v = cf->fmax;

    logError(0, "Value must be at most %*.*f\7",
	    	    cf->width, cf->precision, v);
}


int valid_field(struct FIELDREC *cf, int ret)
{
    int s, c, i, period, chars, btrail, len, tmplen;
    char *p, *q;
    long cl;
    double v, magnitude;
    struct CHOICEREC *cr;

    /*
     * For string data, simply copy out data, advance field, and return.
     * If a string field is required, it must be non-blank.  Hence, the
     * "have_entry" flag is set if a single non-blank character is found.
     * Data is returned left justified, blank padded to end of field.
     */
    switch (cf->dtype) {
      case Char:
	if (ret) {
            p = cf->fdata;
            while (*p && *p == ' ') p++;	    /* skip leading blanks */
            if (*p) cf->have_entry = TRUE;
            else cf->have_entry = FALSE;

            strcpy(cf->argp, p);

            /* Now pad out return value with trailing blanks */
            p = (char *)cf->argp;
            for (i=strlen(p); i<cf->width; i++) p[i] = ' ';
            p[i] = '\0';
	}
	return 1;

      case List:
	if (!ret) return 1;
	if (cf->fdata[0] == '\0' || cf->fdata[0] == ' ') {
	    fbadlist(cf);
	    return 0;
	}
	
	/* Determine true length of field, without trailing 'fillc' */
	len = strlen(cf->fdata);
	p = &cf->fdata[len-1];
	while (*p == cf->fillc && p > cf->fdata) { p--; len--; }
	
	cr=cf->choice_list;
	for (i=0; i<cf->num_choices; i++, cr=cr->next) {
	    if (!strncmp(cf->fdata, cr->choice,
			 max(len, strlen(cr->choice)))) {
		*(int *)cf->argp = cr->ord;
		cf->have_entry = TRUE;
		return(1);
	    }
	}
	fbadlist(cf);
	return 0;

      case Dir:
	if (!ret) return 1;

	if (cf->fdata[0] == '\0' || cf->fdata[0] == ' ') {
	    /* Blank out the receiving argument */
	    memset(cf->argp, ' ', cf->width);
	    *((char *)cf->argp + cf->width) = '\0';
	    cf->have_entry = FALSE;
	    /* Blank field OK here. If it is specified
	     * as mandatory, forminput will catch it
	     */
	    return 1;
	}
	
	len = strlen(cf->fdata);
	p = &cf->fdata[len-1];
	while (*p == cf->fillc && p > cf->fdata) { p--; len--; }
	
	for (i=0, cr=cf->choice_list; i<cf->num_choices; i++, cr=cr->next) {
	    tmplen = max(len, strlen(cr->choice));
	    if (!strncmp(cf->fdata, cr->choice, tmplen)) {
		strcpy(cf->argp, cr->choice);
		strcat(cf->argp, cf->file_ext);
		cf->have_entry = TRUE;
		return 1;
	    }
	}
	fbadlist(cf);
	return 0;

      case Str:
	len = cf->width;
	p = cf->fdata;
	q = cf->argp;
	while (*p && !isspace(*p) && (len > 0)) {
	    *q++ = *p++;
	    len--;
	}
	*q = 0;
	cf->have_entry = (p != cf->fdata);
	return 1;

      default: break;
    }

    /*
     * Syntax scan: There can be at most one leading '+' or '-', and 
     * one period.  Leading or trailing blanks are allowed - blanks 
     * in the middle are not.
     */
    period = chars = btrail = FALSE;
    for (i=0; i<cf->width; i++) {
	c = cf->fdata[i];
	if (btrail && c != ' ') {
            reset_attr(0);
	    logError(0, "Field can not have embedded blanks\7", 0);
            reset_attr(cf->smode);
            return 0;
	} else if (c == '+') {
            if (chars || cf->dtype == Hex) {
                fbadchar(cf, i);
                return 0;
            }
            chars = TRUE;
            cf->fdata[i] = ' ';
	} else if (c == '-') {
            if (chars || cf->dtype == Hex) {
                fbadchar(cf, i);
                return 0;
            }
            chars = TRUE;
	} else if (c == '.') {
            if (period || cf->dtype == Int 
                || cf->dtype == Long || cf->dtype == Hex) {
                fbadchar(cf, i);
                return 0;
            }
            period = chars = TRUE;
	} else if (c == ' ') {
            if (chars) btrail = TRUE;
	} else if ((cf->dtype != Hex && !isdigit(c)) || !isxdigit(c)) {
            fbadchar(cf, i);
            return 0;
	} else chars = TRUE;
    }

    if (!chars) {
	cf->have_entry = FALSE;
        
	/*
         * For non-required fields that are left blank,
	 * return BLANK_INT or BLANK_FLOAT to indicate 'blank field'.
         */
	if (ret && !cf->required) {
	    switch (cf->dtype) {
		case Hex:
		case Int:
                    *(int *)cf->argp = BLANK_INT;
                    break;

		case Long:
                    *(long *)cf->argp = BLANK_LONG;
                    break;
                
		case Float:
                    *(float *)cf->argp = BLANK_FLOAT;
                    break;

		case Double:
                    *(double *)cf->argp = BLANK_DOUBLE;
                    break;

                /* These are needed to keep ansi from complaining. */
	        default: break;
	    }
	}
	return 1;
    }


    switch (cf->dtype) {
    	case Int:	
            s = sscanf(cf->fdata, " %d", &c);
            if (s != 1) {
                fbadchar(cf, 0);
                return 0;
            }
            if (cf->dmin > c) {
                reset_attr(0);
		badmin_int(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (cf->dmax < c) {
                reset_attr(0);
		badmax_int(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (ret) {
                *(int *)cf->argp = c;
                cf->have_entry = TRUE;
            }
            break;
		
	case Long:	
            s = sscanf(cf->fdata, " %ld", &cl);
            if (s != 1) {
                fbadchar(cf, 0);
                return 0;
            }
            if (cf->dmin > cl) {
                reset_attr(0);
		badmin_int(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (cf->dmax < cl) {
                reset_attr(0);
		badmax_int(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (ret) {
                *(long *)cf->argp = cl;
                cf->have_entry = TRUE;
            }
            break;
        
    	case Hex:
            s = sscanf(cf->fdata, " %x", &c);
            if (s != 1) {
                fbadchar(cf, 0);
                return 0;
            }
            if ((unsigned int)cf->dmin > c) {
                reset_attr(0);
		badmin_int(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if ((unsigned int)cf->dmax < c) {
                reset_attr(0);
		badmax_int(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (ret) {
                *(unsigned int *)cf->argp = c;
                cf->have_entry = TRUE;
            }
            break;

	case Float:
            s = sscanf(cf->fdata, " %lf", &v);
            if (s != 1) {
                fbadchar(cf, 0);
                return 0;
            }

	    magnitude = pow(10.0, (double)cf->precision);
        
            if (irint(cf->fmin*magnitude) > irint(v*magnitude)) {
                reset_attr(0);
		badmin_float(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (irint(cf->fmax*magnitude) < irint(v*magnitude)) {
                reset_attr(0);
		badmax_float(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (ret) {
                if (cf->dtype == Float) *(float *)cf->argp = v;
                
                /* must be coord */
                else *(int *)cf->argp = v;

                cf->have_entry = TRUE;
            }
            break;

	case Double:
            s = sscanf(cf->fdata, " %lf", &v);
            if (s != 1) {
                fbadchar(cf, 0);
                return 0;
            }
            if (cf->fmin > v) {
                reset_attr(0);
		badmin_float(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (cf->fmax < v) {
                reset_attr(0);
		badmax_float(cf);
                reset_attr(cf->smode);
                return 0;
            }
            if (ret) {
                *(double *)cf->argp = v;
                cf->have_entry = TRUE;
            }
            break;

        /* These are needed to keep ansi from complaining. */
        default: break;
    }

    return 1;
}
